#!/usr/bin/env python3
# coding=UTF-8
#
# Copyright 2022. quinn.7@foxmail.com All rights reserved.
# Author :: cat7
# Email  :: quinn.7@foxmail.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, 
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
"""
[ 配置 ]

"""

import yaml

from rains.baseic.const import CONST
from rains.baseic.keyword import KEYWORD
from rains.baseic.decorator import singleton_pattern


# ----------------------------------
@singleton_pattern
class Config(object):
    """
    [ 配置 ]

    """

    _data: dict
    """ 配置文件数据 """

    _debug: bool
    """ 调试模式 """

    _driver_max_count: int
    """ 驱动队列最大值 """

    _task_max_count: int
    """ 任务队列最大值 """

    # ------------------------------
    def __init__(self):
        """
        [ 静态配置 ]

        """
        
        # 读取配置文件    
        self.load()

    # ------------------------------
    @property
    def debug(self):
        """
        [ 调试模式 ]

        """

        return self._debug

    # ------------------------------
    @property
    def driver_max_count(self):
        """
        [ 核心队列最大值 ]

        """

        return self._driver_max_count

    # ------------------------------
    @property
    def task_max_count(self):
        """
        [ 任务队列最大值 ]

        """

        return self._task_max_count

    # ------------------------------
    def load(self):
        """
        [ 读取配置 ]

        """

        with open(CONST.SYS.PATH_CONFIG_FILE, 'r', encoding='UTF-8') as f:
            self._data = yaml.load(f, Loader=yaml.FullLoader)
            self._debug = self._data[KEYWORD.CONFIG.DEBUG]
            self._driver_max_count = self._data[KEYWORD.CONFIG.DRIVER_MAX_COUNT]
            self._task_max_count = self._data[KEYWORD.CONFIG.TASK_MAX_COUNT]

    # ------------------------------
    def dump(self):
        """
        [ 存储配置 ]

        """

        with open(CONST.SYS.PATH_CONFIG_FILE, 'w', encoding='UTF-8') as f:
            yaml.dump(self._data, f, allow_unicode=True)

    # ------------------------------
    def set_debug(self, state: bool):
        """
        [ 设置调试模式 ]
        
        ---
        参数:
            state { bool } : 调试模式.

        """

        self._debug = state
        self._data[KEYWORD.CONFIG.DEBUG] = state

    # ------------------------------
    def set_core_max_count(self, number: int):
        """
        [ 设置核心队列最大值 ]
        
        ---
        参数:
            number { int } : 数值.

        """

        self._driver_max_count = number
        self._data[KEYWORD.CONFIG.DRIVER_MAX_COUNT] = number

    # ------------------------------
    def set_task_max_count(self, number: int):
        """
        [ 设置任务队列最大值 ]
        
        ---
        参数:
            number { int } : 数值.

        """

        self._task_max_count = number
        self._data[KEYWORD.CONFIG.TASK_MAX_COUNT] = number


CONFIG: Config = Config()
""" [ 静态配置实例 ] """



#!/usr/bin/env python3
# coding=UTF-8

import setuptools

from rains import __version__
from rains import __author__
from rains import __email__


with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()


setuptools.setup(

    name='rains', 
    version=__version__, 
    author=__author__, 
    author_email=__email__, 
    description='这是一个开箱即用、稳定且高效的工程化全栈自动化测试框架。', 
    long_description=long_description, 
    long_description_content_type='text/markdown', 
    url='https://gitee.com/catcat7/rains', 
    packages=setuptools.find_packages(), 

    install_requires=[
        'Selenium == 3.14.1', 
        'requests', 
        'flask', 
    ], 

    classifiers=[
        'Programming Language :: Python :: 3', 
        'License :: OSI Approved :: MIT License', 
        'Operating System :: OS Independent', 
    ],

    entry_points={
        'console_scripts':[
            'rains = rains.console_rains:main'
        ],
    },
)



#!/usr/bin/env python3
# coding=UTF-8
#
# Copyright 2022. quinn.7@foxmail.com All rights reserved.
# Author :: cat7
# Email  :: quinn.7@foxmail.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, 
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
"""
[ Rains.Core.Common.TaskExecutor ]

"""

import time
from typing import Any

from rains.baseic.log import LOG
from rains.baseic.const import CONST
from rains.baseic.config import CONFIG
from rains.baseic.keyword import KEYWORD
from rains.core.common.action_recorder import ACTION_RECORDER
from rains.core.common.task_fn_agency import TaskFnAgency
from rains.core.common.task_diplomatist import TaskDiplomatist
from rains.core.common.task_instruction import TaskInstruction


# ----------------------------------
class TaskExecutor(object):
    """
    [ 任务执行器 ]

    * 任务的消费者, 它负责接收任务, 并解析运行.

    """

    _driver: Any
    """ [ 套件驱动程序 ] """

    _task_fn_agency: TaskFnAgency
    """ [ 函数代理 ] """

    _task_diplomatist: TaskDiplomatist
    """ [ 任务外交代理 ] """
    
    _task_instruction: TaskInstruction
    """ [ 任务指令 ] """

    # ------------------------------
    def __init__(self, driver: Any):
        """
        [ 任务执行器 ]

        ---
        参数:
            driver_id { int } : 核心ID.
            task_type { str } : 任务类型.
            driver { Any } : 套件驱动程序.
            plant { Any } : 套件构建工厂.

        """

        self._driver = driver
        self._task_fn_agency = TaskFnAgency(driver.plant)
        self._task_diplomatist = TaskDiplomatist(driver.driver_id, driver.project_type)

    # ------------------------------
    def running(self, task: Any):
        """
        [ 开始任务 ]

        * 接收一个 task, 并执行该 task.

        ---
        参数:
            task { Any } : 任务对象, 可以是实现 ITask 接口的任务类, 或者是 JSON 串译文.

        """
        
        self._task_instruction = TaskInstruction()

        if not isinstance(task, dict):
            self._task_instruction.convert_code_task(task)
            self._consumption_task_code()
            
        else:
            self._task_instruction.convert_json_task(task)
            self._consumption_task_json()

    # ------------------------------
    def _consumption_task_code(self):
        """
        [ 执行 CODE 任务 ]
        
        """

        task_map_living: Any
        try:
            task_map_living = self._task_instruction.task_class()
        except BaseException as error:
            raise Exception(error)
        task_map_living.plant = self._driver.plant
        task_state = CONST.STATE.END

        self._driver.running()
        
        try:
            # 创建任务记录
            driver_id = self._driver.driver_id
            task_name = self._task_instruction.task_name
            self._task_diplomatist.created_task(task_name, self._task_instruction.task_remark)
            LOG.info(f'Driver[{ driver_id }] >> 开始任务 >> [ { task_name } ]')

            # 执行任务初始化
            self._running_code_setter_instruction(KEYWORD.TASK.SET_TASK_INIT, task_map_living)

            # 执行步骤指令集
            for case_name, case_function in self._task_instruction.case_instructions.items():

                # 执行用例初始化
                self._running_code_setter_instruction(KEYWORD.TASK.SET_CASE_INIT, task_map_living)

                # 执行用例
                self._execute_code_case(case_name, task_map_living, case_function)

                # 执行任务注销
                self._running_code_setter_instruction(KEYWORD.TASK.SET_CASE_QUIT, task_map_living)

                # 引擎重置
                self._driver.reset()

            # 执行任务注销
            self._running_code_setter_instruction(KEYWORD.TASK.SET_TASK_QUIT, task_map_living)

        except BaseException as error:
            task_state = CONST.STATE.ERR
            LOG.info(f'Driver[{ driver_id }] >> 由于未知异常任务被迫中止:: { error }')
        
        finally:
            self._driver.end()
            
            # 更新数据库任务记录
            self._task_diplomatist.end(task_state)
            LOG.info(f'Driver[{ driver_id }] >> 完成任务 >> [ { task_name } ]')

    # ------------------------------
    def _running_code_setter_instruction(self, setter: str, living: Any):
        """
        [ 执行 CODE 设置器指令 ]
        
        """
        
        if self._task_instruction.setter_instructions[setter]:
            self._task_instruction.setter_instructions[setter](living)

    # ------------------------------
    def _execute_code_case(self, case_name: str, task_map_living: Any, case_function: Any):
        """
        [ 执行 CODE 任务用例 ]

        ---
        参数:
            case_name { str } : 用例名称.
            task_map_living { Any } : 任务类实例.
            case_function { Any } : 用例函数对象.

        """

        # 创建用例记录
        case_remark = 'NULL'
        if case_function.__doc__:
            case_remark = case_function.__doc__.strip()
        case_id = self._task_diplomatist.created_case(case_name, case_remark)
        
        case_state: str
        case_result: bool
        case_start_time = time.strftime("%H:%M:%S")
        case_start_timestamp = int(time.time())
        try:
            LOG.info(f'Driver[{ self._driver.driver_id }] >> 开始用例 >> [ { case_name } ]')

            # 执行用例
            result = case_function(task_map_living)

            # 如果用例返回 True 或者未返回 False 都判断为用例通过
            if result is True or result is None:
                case_state = CONST.STATE.END
                case_result = True

            # 如果用例返回 False 则判断用例失败
            else:
                case_state = CONST.STATE.ERR
                case_result = False

        # 如果用例执行过程发生异常则判断用例失败
        except BaseException as error:
            if CONFIG.debug:
                LOG.info(f'用例执行失败 : { error }')
            case_state = CONST.STATE.ERR
            case_result = False

        # 更新用例记录
        finally:
            LOG.info(f'Driver[{ self._driver.driver_id }] >> 完成用例 >> [ { case_name } ] >> [ { case_result } ]')
            ACTION_RECORDER.write(self._driver.driver_id, f'用例执行结果 >> [{ case_result }]')

            self._task_diplomatist.update_case(
                case_id, 
                case_state, 
                case_start_time, 
                case_start_timestamp
            )

    # ------------------------------
    def _consumption_task_json(self):
        """
        [ 执行译文任务 ]
        
        """
        
        task_state = CONST.STATE.END

        # 运行引擎
        self._driver.running()

        try:
            # 创建任务记录
            driver_id = self._driver.driver_id
            task_name = self._task_instruction.task_name
            self._task_diplomatist.created_task(task_name, self._task_instruction.task_remark)
            LOG.info(f'Driver[{ driver_id }] >> 开始任务 >> [ { task_name } ]')

            # 执行任务初始化
            self._running_json_setter_instruction(KEYWORD.TASK.SET_TASK_INIT)

            # 执行步骤指令集
            for case_name, case_dict in self._task_instruction.case_instructions.items():

                # 执行用例初始化
                self._running_json_setter_instruction(KEYWORD.TASK.SET_CASE_INIT)

                # 执行用例
                self._execute_json_case(case_name, case_dict)

                # 执行用例注销
                self._running_json_setter_instruction(KEYWORD.TASK.SET_CASE_QUIT)
                
                # 引擎重置
                self._driver.reset()

            # 执行任务注销
            self._running_json_setter_instruction(KEYWORD.TASK.SET_TASK_QUIT)

        except BaseException as error:
            task_state = CONST.STATE.ERR
            LOG.info(f'Driver[{ driver_id }] >> 由于未知异常任务被迫中止:: { error }')
        
        finally:
            self._driver.end()
            
            # 更新数据库任务记录
            self._task_diplomatist.end(task_state)
            LOG.info(f'Driver[{ driver_id }] >> 完成任务 >> [ { task_name } ]')

    # ------------------------------
    def _running_json_setter_instruction(self, setter: str):
        """
        [ 执行 JSON 设置器指令 ]
        
        """
        
        if self._task_instruction.setter_instructions[setter]:
            self._task_fn_agency.map(self._task_instruction.setter_instructions[setter])

    # ------------------------------
    def _execute_json_case(self, case_name: str, case_dict: dict):
        """
        [ 执行 JSON 用例 ]

        ---
        参数:
            case_name { str } : 用例名称.
            case_dict { dict } : 用例字典.

        """

        # 创建用例记录
        case_remark = 'NULL'
        if not case_dict['case_remark']:
            case_remark = case_dict['case_remark'].strip()

        case_id = self._task_diplomatist.created_case(case_name, case_remark)
        
        case_state: str
        case_result: bool
        case_start_time = time.strftime("%H:%M:%S")
        case_start_timestamp = int(time.time())

        step_return_dict: dict = {}
        try:
            # 执行用例
            for k, v in case_dict['step_list'].items():
                r = self._task_fn_agency.map(v)
                step_return_dict.update({k: r})
            
            if len(case_dict['assert']) > 0:
                for k, v in case_dict['assert'].items():
                    if step_return_dict[k] == v:
                        case_state = CONST.STATE.END
                        case_result = True
                    else:
                        case_state = CONST.STATE.ERR
                        case_result = False
            else:
                case_state = CONST.STATE.END
                case_result = True

        # 如果用例执行过程发生异常则判断用例失败
        except BaseException as err:
            LOG.debug(f'{ err }')
            case_state = CONST.STATE.ERR
            case_result = False

        # 更新用例记录
        finally:
            LOG.info(f'Driver[{ self._driver.driver_id }] >> 完成用例 >> [ { case_name } ] >> [ { case_result } ]')
            ACTION_RECORDER.write(self._driver.driver_id, f'用例执行结果 >> [{ case_result }]')

            self._task_diplomatist.update_case(
                case_id, 
                case_state, 
                case_start_time, 
                case_start_timestamp
            )



#!/usr/bin/env python3
# coding=UTF-8
#
# Copyright 2022. quinn.7@foxmail.com All rights reserved.
# Author :: cat7
# Email  :: quinn.7@foxmail.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, 
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
"""
[ Rains.DB.RainsSql ]

"""

from rains.server.db.blueprint.sql_task import SqlTask
from rains.server.db.blueprint.sql_case import SqlCase


# ----------------------------------
class RainsSql(object):
    """
    [ SQL语句映射类 ]

    """

    @property
    def task(self) -> SqlTask:
        """
        [ 任务相关 ]

        ---
        返回:
            SqlTask : SQL语句任务相关封装类.

        """

        return SqlTask

    # ------------------------------
    @property
    def case(self) -> SqlCase:
        """
        [ 用例相关 ]

        ---
        返回:
            SqlCase : SQL语句用例相关封装类.

        """

        return SqlCase



#!/usr/bin/env python3
# coding=UTF-8
#
# Copyright 2022. quinn.7@foxmail.com All rights reserved.
# Author :: cat7
# Email  :: quinn.7@foxmail.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, 
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
"""
[ 任务关键字 ]

"""


# ----------------------------------
class KeywordTask(object):
    """ [ 任务关键字 ] """

    SET_TASK_INIT = 'set_task_init'
    """ [ 任务初始化 ] """

    SET_TASK_QUIT = 'set_task_quit'
    """ [ 任务注销 ] """

    SET_CASE_INIT = 'set_case_init'
    """ [ 用例初始化 ] """

    SET_CASE_QUIT = 'set_case_quit'
    """ [ 用例注销 ] """

    JSON_ARG_MODULE = 'module'
    """ [ JSON字段::模块 ] """

    JSON_ARG_CLASS = 'class'
    """ [ JSON字段::类别 ] """

    JSON_ARG_FUNCTION = 'function'
    """ [ JSON字段::方法 ] """

    JSON_ARG_ELEMENT = 'element'
    """ [ JSON字段::元素 ] """

    JSON_ARG_PARAMETER = 'parameter'
    """ [ JSON字段::参数 ] """



#!/usr/bin/env python3
# coding=UTF-8
#
# Copyright 2022. quinn.7@foxmail.com All rights reserved.
# Author :: cat7
# Email  :: quinn.7@foxmail.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, 
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
"""
[ 系统常量 ]

"""

import pathlib
from pathlib import Path


# ----------------------------------
class ConstSys(object):
    """ [ 系统常量 ] """

    PATH_ROOT: Path = pathlib.Path().cwd()
    """ [ 运行路径 ] """

    NAME_DATA_DIR: str = 'data'
    """ [ 数据文件夹名称 ] """

    PATH_DATA_DIR: Path = PATH_ROOT.joinpath(NAME_DATA_DIR)
    """ [ 数据文件夹路径 ] """

    NAME_LOGS_DIR: str = 'log'
    """ [ 日志文件夹名称 ] """

    PATH_LOGS_DIR: Path = PATH_DATA_DIR.joinpath(NAME_LOGS_DIR)
    """ [ 日志文件夹路径 ] """

    NAME_DB_FILE: str = 'RainsDatabase.db'
    """ [ 数据库文件名称 ] """

    PATH_DB_FILE: Path = PATH_DATA_DIR.joinpath(NAME_DB_FILE)
    """ [ 项目数据库文件 ] """

    NAME_CONFIG_FILE: str = 'RainsConfig.yaml'
    """ [ 配置文件名称 ] """

    PATH_CONFIG_FILE: Path = PATH_DATA_DIR.joinpath(NAME_CONFIG_FILE)
    """ [ 配置文件路径 ] """

    NAME_RAINS_INI_FILE: str = 'rains.ini'
    """ [ INI文件名称 ] """

    PATH_RAINS_INI_FILE: Path = PATH_ROOT.joinpath(NAME_RAINS_INI_FILE)
    """ [ INI文件路径 ] """



#!/usr/bin/env python3
# coding=UTF-8
#
# Copyright 2022. quinn.7@foxmail.com All rights reserved.
# Author :: cat7
# Email  :: quinn.7@foxmail.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, 
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
"""
[ 创建任务 ]

"""

import os
import sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__)))))

import re
from pathlib import Path

from rains.baseic.const import CONST


# ----------------------------------
def function_create_task():
    """
    [ 创建任务 ]

    """

    if len(sys.argv) >= 4:
        # 获取任务类型与模板路径
        task_template_path: Path
        task_type: str = sys.argv[2]
        if task_type in ['web', 'Web', 'WEB', 'w', 'W']:
            task_template_path = Path(sys.argv[0]).parent.joinpath(TASK_TEMPLATE_WEB_FILE_PATH)
        elif task_type in ['api', 'Api', 'API', 'a', 'A']:
            task_template_path = Path(sys.argv[0]).parent.joinpath(TASK_TEMPLATE_API_FILE_PATH)
        else:
            print(_MESSAGE_ERROR)
            exit(0)

        print('--------------------------------------')
        for task_name in sys.argv[3:]:
            # 拼接合法的任务名称
            if len(task_name) < 5 or task_name[0:5] != TASK_NAMING_BASE_PREFIX:
                task_name = ''.join([TASK_NAMING_BASE_PREFIX, task_name])
            if '.py' not in task_name:
                task_name2 = ''.join([task_name, '.py'])
            print(f'任务 [{ task_name }] => ', end='')
            
            task_path: Path = CONST.SYS.PATH_ROOT.joinpath(task_name2)
            if not task_path.exists():
                # 拼接首字母大写的任务类名
                template_class_name = ''
                for i in task_name.split('_'):
                    template_class_name = ''.join([template_class_name, i.capitalize()]) 
                
                # 创建任务
                task_template_code = re.sub(r'class Task', f'class { template_class_name }', task_template_path.read_text(encoding='UTF-8'))
                task_path.touch(mode=0o777)
                task_path.write_bytes(bytes(task_template_code, encoding='UTF-8'))
                print(_MESSAGE_CREATING_SUCCESSFUL)
            else:
                print(_MESSAGE_CREATING_UNSUCCESSFUL)
        print('--------------------------------------')

    else:
        print(_MESSAGE_ERROR)


# 任务命名前缀
TASK_NAMING_BASE_PREFIX: str = 'task_'

# Web任务模板路径
TASK_TEMPLATE_WEB_FILE_PATH: str = 'task_template\\task_web_template.py'

# Api任务模板路径
TASK_TEMPLATE_API_FILE_PATH: str = 'task_template\\task_api_template.py'

# 任务创建成功文案
_MESSAGE_CREATING_SUCCESSFUL: str = '创建成功!'

# 任务创建失败文案
_MESSAGE_CREATING_UNSUCCESSFUL: str = '创建失败, 任务已经存在!'

# 任务创建异常文案
_MESSAGE_ERROR: str = \
"""
--------------------------------------

指令键入异常!

正确格式:
>> rains -task [task_type] [task_name]
>> rains -task [task_type] [task_name] [task_name] ...

示例:
>> rains -task task_t1
>> rains -task task_t1 task_t2

目前支持的任务类型:
web : Web功能测试任务
api : Api功能测试任务

--------------------------------------
"""



#!/usr/bin/env python3
# coding=UTF-8
#
# Copyright 2022. quinn.7@foxmail.com All rights reserved.
# Author :: cat7
# Email  :: quinn.7@foxmail.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, 
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
"""
[ Rains.Kit.Web.Packaging.WebFindHandle ]

"""

from typing import Any

from selenium.common.exceptions import TimeoutException
from selenium.webdriver.support import expected_conditions as ec
from selenium.webdriver.support.wait import WebDriverWait
from selenium.webdriver.chrome.webdriver import RemoteWebDriver

from rains.baseic.error import WebFindHandleFindElementError
from rains.kit.web.packaging.web_element_structure import WebElementStructure


# ----------------------------------
class WebFindHandle(object):
    """
    [ Web元素定位器 ]

    """

    _driver: RemoteWebDriver
    """ [ 原生驱动对象 ] """

    _web_driver_wait: WebDriverWait = WebDriverWait
    """ [ 原生驱动显示等待函数 ] """

    # ------------------------------
    def __init__(self, base_driver: RemoteWebDriver):
        """
        [ Web元素定位器 ]

        ---
        参数:
            base_driver { RemoteWebDriver } : Selenium 原生驱动对象.

        """

        self._driver = base_driver

    # ------------------------------
    def find(self, 
        element_structure: WebElementStructure, 
        wait: int = 15, 
        tolerance: bool = False
    ) -> list:
        """
        [ 元素定位 ]

        * 接受一个 WebElementStructure 对象, 并根据其描述从当前页面上定位元素列表.

        ---
        参数:
            element_structure { WebElementStructure } : 元素结构体.
            wait { int, optional } : 查找元素所等待的显式等待时长.
            tolerance { bool, optional } : 容错模式, 为 True 则表示定位不到元素也不报错, 返回空列表.

        ---
        返回:
            list : 已获取定位的元素列表.

        """

        # 解析元素搜索范围
        search_range = self._analysis_search_range(element_structure, wait)

        # 开始定位UI元素
        return self._base_explicitly_wait_find_element(
            by_key=element_structure.by_key, 
            by_value=element_structure.by_value, 
            driver=search_range, 
            wait=wait, 
            tolerance=tolerance
        )

    # ------------------------------
    def _analysis_search_range(self, element_structure: WebElementStructure, wait: int) -> Any:
        """
        [ 解析元素搜索范围 ]

        * 如果元素结构体中携带了锚元素则会返回定位的锚元素对象, 否则返回驱动. 

        ---
        参数:
            element_structure { WebElementStructure } : 元素结构体.
            wait { int } : 查找元素所等待的显式等待时长.

        ---
        返回:
            Any : 驱动 or 锚元素对象.

        """

        search_range = self._driver

        if element_structure.anchor_by_key:
            search_range = self._base_explicitly_wait_find_element(
                by_key=element_structure.anchor_by_key, 
                by_value=element_structure.anchor_by_value, 
                driver=search_range, 
                wait=wait, 
                tolerance=False
            )[element_structure.anchor_location_id]

        return search_range

    # ------------------------------
    def _base_explicitly_wait_find_element(self, 
                                           by_key: str, 
                                           by_value: str, 
                                           driver: RemoteWebDriver, 
                                           wait: int, 
                                           tolerance: bool) -> list:
        """
        [ 原生 WebDriverWait 封装 ]

        * 封装 WebDriverWait 全局显示等待获取函数.
        * 该函数将在每 0.5 秒时遍历一次页面, 直到找到元素或者等待(wait)超时. 
        * 如果容错开关(tolerance)为 True, 则表示定位不到元素时返回空列表.

        ---
        参数:
            by_key { str } : 元素定位策略.
            by_value { str } : 元素定位策略对应的值.
            driver { RemoteWebDriver } : Selenium 原生驱动对象.
            wait { int } : 查找元素所等待的显式等待时长.
            tolerance { bool } : 容错模式, 为 True 则表示定位不到元素也不报错, 返回空列表.

        ---
        异常:
            TimeoutException : 获取元素超时异常.
            WebFindHandleFindElementError : Web定位器获取元素异常.

        ---
        返回:
            list : 已获取定位的元素列表 or 空列表.

        """

        try:
            return self._web_driver_wait(driver, wait, poll_frequency=0.5, ignored_exceptions=None)\
                .until(ec.presence_of_all_elements_located((by_key, by_value)))

        except TimeoutException as error:
            # 触发异常时, 如果容错开关(tolerance)为 True, 则不抛出异常, 仅返回空列表
            if tolerance:
                return []
            else:
                raise WebFindHandleFindElementError(error)



#!/usr/bin/env python3
# coding=UTF-8
#
# Copyright 2022. quinn.7@foxmail.com All rights reserved.
# Author :: cat7
# Email  :: quinn.7@foxmail.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, 
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
"""
[ Rains.Kit.Api.ApiDriver ]

"""

import random

from rains.baseic.const import CONST
from rains.baseic.interface.i_driver import IDriver
from rains.core.common.task_executor import TaskExecutor

from rains.kit.api.api_plant import ApiPlant


API_DRIVER_SIGN_ID_LIST = []
""" [ 驱动程序编号池 ] """


# ----------------------------------
class ApiDriver(IDriver):
    """
    [ Api 驱动程序 ]

    """

    _driver_id: int
    """ [ 驱动程序ID ] """

    _state: bool = False
    """ [ 运行状态 ] """
    
    _plant: ApiPlant = None
    """ [ 构建工厂对象 ] """

    _task_executor: TaskExecutor
    """ [ 任务执行器 ] """

    _project_type: str = CONST.TYPE.PROJECT_TYPE_API
    """ [ 项目类型 ] """

    _headers: dict = {
        'Authorization': 'NULL',
        'User-Agent'   : 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.11; rv:46.0) Gecko/20100101 Firefox/46.0',
    }
    """ [ 头部信息 ] """

    # ------------------------------
    def __init__(self):
        """
        [ Web 驱动程序 ]

        """

        # 生成随机ID并创建驱动程序
        while True:
            self._driver_id = random.randint(100, 999)

            if self._driver_id not in API_DRIVER_SIGN_ID_LIST:
                API_DRIVER_SIGN_ID_LIST.append(self._driver_id)
                break
            else:
                continue
            
        self._plant = ApiPlant(self)
        self._task_executor = TaskExecutor(self)

    # ------------------------------
    @property
    def plant(self):
        """
        [ 构建工厂 ]

        """

        return self._plant

    # ------------------------------
    @property
    def driver_id(self):
        """
        [ 驱动ID ]

        """

        return self._driver_id

    # ------------------------------
    @property
    def project_type(self):
        """
        [ 项目类型 ]

        """

        return self._project_type

    # ------------------------------
    @property
    def headers(self):
        """
        [ 头部信息 ]

        """

        return self._headers

    # ------------------------------
    def set_headers(self, new_headers: dict):
        """
        [ 设置头部信息 ]

        ---
        参数:
            new_headers { dict } : 新的请求头字典.

        """
        
        self._headers = new_headers

    # ------------------------------
    def get_state(self) -> bool:
        """
        [ 获取引擎运行状态 ]

        * 获取引擎运行状态, 启动中返回 True, 否则返回 False. 

        Returns:
            bool: [ 引擎运行状态 ]

        """

        return self._state

    # ------------------------------
    def start_task(self, task):
        """
        [ 开始任务 ]

        * 接收一个 task, 并执行该 task.

        ---
        参数:
            task { Any } : 任务对象, 可以是实现 ITask 接口的任务类, 或者是 JSON 串译文.

        """

        self._task_executor.running(task)

    # ------------------------------
    def running(self):
        """
        [ 运行引擎 ]

        * 驱动与定位器的实例化.

        """

        if not self._state:
            self._state = True

    # ------------------------------
    def end(self):
        """
        [ 结束引擎 ]

        * 退出并销毁当前持有的驱动与定位器. 

        """

        if self._state:
            self._state = False

    # ------------------------------
    def reset(self):
        """
        [ 重置引擎 ]

        * 将引擎重置至初始阶段.
        * 与 end() 相比, 该函数只是重置引擎至初始状态, 而不是注销.

        """

        if self._state:
            ...



#!/usr/bin/env python3
# coding=UTF-8
#
# Copyright 2022. quinn.7@foxmail.com All rights reserved.
# Author :: cat7
# Email  :: quinn.7@foxmail.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, 
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
"""
[ 指令集 ]

"""

import os
import sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(os.path.realpath(__file__)))))


_MESSAGE = \
"""
--------------------------------------
 rains
--------------------------------------

 -init    | -i :: 初始化工程
 -make    | -m :: 创建项目
 -task    | -t :: 创建任务
 -run     | -r :: 运行任务文件或目录
 -server  | -s :: 运行服务器
 -help    | -h :: 查看帮助
 -version | -v :: 查看版本

--------------------------------------
"""


def function_order():
  """
  [ 指令集 ]
  
  """

  print(_MESSAGE)



#!/usr/bin/env python3
# coding=UTF-8
#
# Copyright 2022. quinn.7@foxmail.com All rights reserved.
# Author :: cat7
# Email  :: quinn.7@foxmail.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, 
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
"""
[ Rains.Common.Task.TaskInstruction ]

"""

from typing import Any

from rains.baseic.const import CONST
from rains.baseic.keyword import KEYWORD
from rains.baseic.error import TaskInstructionError


CASE_SPECIFICATION_NAMING = 'CASE'
""" [ 合法用例命名 ] """


# ----------------------------------
class TaskInstruction(object):
    """
    [ 任务指令 ]

    * Rains 允许通过多种方式编写自动化测试用例.
    * 为了使这些方式具备一定的兼容性, 该类将进行中间指令集的转换.

    """

    _task_name: str
    """ [ 任务名称 ] """

    _task_remark: str
    """ [ 任务注释 ] """

    _task_type: str
    """ [ 任务类型 ] """

    _task_class = None
    """ [ 任务类的类名 ] """

    _case_instructions: dict
    """ [ 用例指令集 ] """

    _setter_instructions: dict
    """ [ 设置器指令集 ] """
    
    _is_packaging: bool = False
    """ [ 是否已转换 ] """

    # ------------------------------
    def __init__(self):
        """
        [ 任务指令 ]

        """

        self._case_instructions = {}

        self._setter_instructions = {
            KEYWORD.TASK.SET_TASK_INIT: None, 
            KEYWORD.TASK.SET_TASK_QUIT: None, 
            KEYWORD.TASK.SET_CASE_INIT: None, 
            KEYWORD.TASK.SET_CASE_QUIT: None
        }

    # ------------------------------
    @property
    def task_name(self) -> str:
        """
        [ 任务类型 ]

        """

        return self._task_name

    # ------------------------------
    @property
    def task_remark(self) -> str:
        """
        [ 任务注释 ]

        """

        return self._task_remark

    # ------------------------------
    @property
    def task_type(self) -> str:
        """
        [ 任务类型 ]

        """

        return self._task_type

    # ------------------------------
    @property
    def task_class(self) -> str:
        """
        [ 任务类的类名 ]

        """

        return self._task_class

    # ------------------------------
    @property
    def case_instructions(self) -> dict:
        """
        [ 用例指令集 ]

        """

        return self._case_instructions

    # ------------------------------
    @property
    def setter_instructions(self) -> dict:
        """
        [ 设置器指令集 ]

        """

        return self._setter_instructions

    # ------------------------------
    def convert_code_task(self, task_class: Any):
        """
        [ 转换代码类型任务 ]
        
        ---
        参数:
            task_class { Any } : 继承 IKitTask 套件任务接口的任务类.

        """

        try:
            if not self._is_packaging:

                # 确认 指令集类型
                self._task_type = CONST.TYPE.TASK_EXECUTION_MODE_CODE
                # 获取 任务类的类名
                self._task_class = task_class
                # 读取 任务类的信息
                self._task_name = task_class.__name__
                if task_class.__doc__:
                    self._task_remark = task_class.__doc__.strip()
                else:
                    self._task_remark = 'NULL'

                # 解析指令集
                for k, v in task_class.__dict__.items():

                    # 捕获设置器指令集
                    if k in self._setter_instructions.keys() and callable(v):
                        self._setter_instructions[k] = v

                    # 捕获步骤指令集
                    if str(k[:4:]).upper() == CASE_SPECIFICATION_NAMING and callable(v):
                        self._case_instructions[k] = v

                # 完成封装
                self._is_packaging = True

            return self

        except BaseException as error:
            raise TaskInstructionError(error)


    # ------------------------------
    def convert_json_task(self, task_json: dict):
        """
        [ 转换译文类型任务 ]

        ---
        参数:
            task_json { dict } : 符合规范的 JSON 译文字典.

        """

        try:
            if not self._is_packaging:

                # 确认 指令集类型
                self._task_type = CONST.TYPE.TASK_EXECUTION_MODE_JSON
                # 读取 任务类的信息
                self._task_name = task_json['task_name']
                self._task_remark = task_json['task_remark']

                # 解析设置器指令集
                instructions_setter: dict = task_json['instructions_setter']
                for k, v in instructions_setter.items():
                    if k in self._setter_instructions.keys():
                        self._setter_instructions[k] = v

                # 解析用例指令集
                instructions_case: dict = task_json['instructions_case']
                for k, v in instructions_case.items():
                    if str(k[:4:]).upper() == CASE_SPECIFICATION_NAMING:
                        self._case_instructions[k] = v

                # 完成封装
                self._is_packaging = True

            return self

        except BaseException as error:
            raise TaskInstructionError(error)



#!/usr/bin/env python3
# coding=UTF-8
#
# Copyright 2022. quinn.7@foxmail.com All rights reserved.
# Author :: cat7
# Email  :: quinn.7@foxmail.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, 
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
"""
[ Web元素结构体 ]

"""


# ----------------------------------
class WebElementStructure(object):
    """
    [ Web元素结构体 ]


    """

    page: str
    """ [ 页面名称 ] """

    name: str
    """ [ 组件名称 ] """

    by_key: str
    """ [ 元素定位策略 ] """

    by_value: str
    """ [ 元素定位策略对应的值 ] """

    anchor_by_key: str
    """ [ 锚点元素定位策略 ] """

    anchor_by_value: str
    """ [ 锚点元素定位策略 ] """

    anchor_location_id: int
    """ [ 锚点元素定位ID ] """

    # ------------------------------
    def __init__(
        self, 
        page: str, 
        name: str, 
        by_key: str, 
        by_value: str, 
        anchor_by_key: str = None, 
        anchor_by_value: str = None, 
        anchor_location_id: int = None
    ):
        """
        [ Web元素结构体 ]

        * 构建组件时, 需要传递组件于页面上的定位方式(by_key)与相应的值(by_value).
        * 当传递了非必须的锚点元素时, 表示将在该锚点节点中定位元素.
        * 如果传递锚点元素(anchor), 则必须同时传递 anchor_by_key、anchor_by_value、anchor_location_id.

        ---
        参数:
            page { str } : 页面名称.
            name { str } : 组件名称.
            by_key { str } : 元素定位策略.
            by_value { str } : 元素定位策略对应的值.
            anchor_by_key { str, optional } : 锚点元素定位策略.
            anchor_by_value { str, optional } : 锚点元素定位策略对应的值.
            anchor_location_id { str, optional } : 锚点元素定位ID.

        """

        # 获取元素参数
        self.page = page
        self.name = name
        self.by_key = by_key
        self.by_value = by_value
        self.anchor_by_key = anchor_by_key
        self.anchor_by_value = anchor_by_value
        self.anchor_location_id = anchor_location_id



#!/usr/bin/env python3
# coding=UTF-8
#
# Copyright 2022. quinn.7@foxmail.com All rights reserved.
# Author :: cat7
# Email  :: quinn.7@foxmail.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, 
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================
"""
[ Rains.Kit.Web.Packaging.WebPlantElement ]

"""

import time
from typing import Any

from selenium.webdriver import ActionChains
from selenium.webdriver.support.select import Select

from rains.baseic.log import LOG
from rains.baseic.config import CONFIG
from rains.core.common.action_recorder import ACTION_RECORDER
from rains.baseic.tools import Regular

from rains.kit.web.packaging.web_element_structure import WebElementStructure


# ----------------------------------
class WebPlantElement(object):
    """
    [ Web元素控件 ]

    * 基于元素特征构建指向该元素的元素控件对象.
    * 封装了对页面元素的操作功能, 如获取元素信息、模拟鼠标、输入框、下拉框等操作.

    """

    _web_driver: Any
    """ 引擎对象 """

    _element_structure: WebElementStructure
    """ 元素结构体 """

    # ------------------------------
    def __init__(self, web_driver: Any, element_structure: WebElementStructure):
        """
        [ Web元素控件 ]

        ---
        参数:
            web_driver { WebDriver } : Web引擎对象.
            element_structure { WebElementStructure } : Web元素结构体.

        """

        self._web_driver = web_driver
        self._element_structure = element_structure

    # ------------------------------
    @property
    def get(self):
        """
        [ Web元素控件'获取'相关函数 ]

        """

        return _WebPlantElementMapGet(self._web_driver, self._element_structure)

    # ------------------------------
    def _map_get(self):
        return self.get

    # ------------------------------
    @property
    def mouse(self):
        """
        [ Web元素控件'鼠标'相关函数 ]

        """

        return _WebPlantElementMapMouse(self._web_driver, self._element_structure)

    # ------------------------------
    def _map_mouse(self):
        return self.mouse

    # ------------------------------
    @property
    def input(self):
        """
        [ Web元素控件'输入框'相关函数 ]

        """

        return _WebPlantElementMapInput(self._web_driver, self._element_structure)

    # ------------------------------
    def _map_input(self):
        return self.input

    # ------------------------------
    @property
    def selector(self):
        """
        [ Web元素控件'下拉框'相关函数 ]

        """

        return _WebPlantElementMapSelector(self._web_driver, self._element_structure)

    # ------------------------------
    def _map_selector(self):
        return self.selector


# ----------------------------------
class _WebPlantElementMapBase(object):
    """
    [ Web元素控件映射基础类 ]

    """

    _web_driver: Any
    _handler: Any
    _driver_id: int
    _element_container: list
    _element_structure: WebElementStructure

    # ------------------------------
    def __init__(self, web_driver: Any, element_structure: WebElementStructure):
        """
        [ Web元素控件映射基础类 ]

        ---
        参数:
            web_driver { WebDriver } : Web引擎对象.
            element_structure { WebElementStructure } : Web元素结构体.

        """

        self._web_driver = web_driver
        self._handler = web_driver.base_driver
        self._driver_id = web_driver.driver_id
        self._element_structure = element_structure

    # ------------------------------
    def _analytical_elements(self, wait: int = 15, tolerance: bool = False) -> list:
        """
        [ 解析元素结构 ]

        * 每一次调用页面元素组件的方法时, 都会调用该函数, 以更新定位结果, 确认组件是否存在于当前页面中.

        ---
        参数:
            wait { int, optional } : 查找元素所等待的显式等待时长.
            tolerance { bool, optional } : 容错模式, 为 True 则表示定位不到元素也不报错, 返回空列表.

        ---
        返回:
            list : 元素列表容器.

        """

        self._element_container = \
            self._web_driver.find_handle.find(self._element_structure, wait, tolerance)

        return self._element_container

    # ------------------------------
    def _prepare(self, message: str, is_inquire: bool = True):
        message = ''.join([
            self._element_structure.page, '.', 
            self._element_structure.name, ' >> ', 
            message
        ])

        if is_inquire:
            self._element_container = self._analytical_elements()

        return message


# ----------------------------------
class _WebPlantElementMapGet(_WebPlantElementMapBase):
    """
    [ Web元素控件'获取'相关函数 ]

    """

    # ------------------------------
    def count(self) -> int:
        """
        [ 获取元素数量 ]

        * 定位元素并且返回元素数量.

        ---
        返回:
            int : 元素数量.

        """

        try:
            message = self._prepare(Regular.get_function_doc_name(self.count.__doc__))
            count = len(self._element_container)
            execution_step_record(self._driver_id, message, count)

            return count

        except BaseException as error:
            raise WebPlantElementFunctionRunError(self._driver_id, message, error)

    # ------------------------------
    def text(self, eid: int = 0) -> str:
        """
        [ 获取文本 ]

        * 返回元素的 text 属性值.

        ---
        参数:
            eid { int, optional } : 指定元素下标, 默认为 0.

        ---
        返回:
            str : 元素文本.

        """

        try:
            message = self._prepare(Regular.get_function_doc_name(self.text.__doc__))
            text = self._element_container[eid].text
            execution_step_record(self._driver_id, message, text)

            return text

        except BaseException as error:
            raise WebPlantElementFunctionRunError(self._driver_id, message, error)

    # ------------------------------
    def value(self, eid: int = 0) -> str:
        """
        [ 获取值 ]

        * 返回元素的 value 属性值.

        ---
        参数:
            eid { int, optional } : 指定元素下标, 默认为 0.

        ---
        返回:
            str : 元素的值.

        """

        try:
            message = self._prepare(Regular.get_function_doc_name(self.value.__doc__))
            value = self._element_container[eid].get_attribute('value')
            execution_step_record(self._driver_id, message, value)

            return value

        except BaseException as error:
            raise WebPlantElementFunctionRunError(self._driver_id, message, error)

    # ------------------------------
    def exist(self, eid: int = 0, wait: int = 3) -> bool:
        """
        [ 检查控件是否存在 ]

        * 尝试在当前页面中定位控件, 在等待时间内未找到则返回 False.

        ---
        参数:
            eid { int, optional } : 指定元素下标, 默认为 0.
            wait { int, optional } : 查找元素所等待的显式等待时长, 默认为 3.

        ---
        返回:
            bool : True or False.

        """

        try:
            message = self._prepare(Regular.get_function_doc_name(self.exist.__doc__), is_inquire=False)
            self._analytical_elements(wait=wait, tolerance=True)

            proof = False
            if len(self._element_container) >= (eid + 1):
                proof = True

            execution_step_record(self._driver_id, message, str(proof))

            return proof

        except BaseException as error:
            raise WebPlantElementFunctionRunError(self._driver_id, message, error)


# ----------------------------------
class _WebPlantElementMapMouse(_WebPlantElementMapBase):
    """
    [ Web元素控件'鼠标'相关函数 ]

    """

    # ------------------------------
    def tap(self, eid: int = 0):
        """
        [ 左键点击 ]

        * 模拟一次作用于 WebElement 对象上的鼠标左键单击操作.

        ---
        参数:
            eid { int, optional } : 指定元素下标, 默认为 0.

        """

        try:
            message = self._prepare(Regular.get_function_doc_name(self.tap.__doc__))
            ActionChains(self._handler).click(self._element_container[eid]).perform()
            execution_step_record(self._driver_id, message)

        except BaseException as error:
            raise WebPlantElementFunctionRunError(self._driver_id, message, error)

    # ------------------------------
    def double_tap(self, eid: int = 0):
        """
        [ 左键双击 ]

        * 模拟一次作用于 WebElement 对象上的鼠标左键双击操作.

        ---
        参数:
            eid { int, optional } : 指定元素下标, 默认为 0.

        """

        try:
            message = self._prepare(Regular.get_function_doc_name(self.double_tap.__doc__))
            ActionChains(self._handler).double_click(self._element_container[eid]).perform()
            execution_step_record(self._driver_id, message)

        except BaseException as error:
            raise WebPlantElementFunctionRunError(self._driver_id, message, error)

    # ------------------------------
    def long_tap(self, eid: int = 0, sleep: int = 2):
        """
        [ 左键长按 ]

        * 模拟一次作用于 WebElement 对象上的鼠标左键长按操作.

        ---
        参数:
            eid { int, optional } : 指定元素下标, 默认为 0.
            sleep { int, optional } : 长按时长, 默认为 2 秒.

        """

        try:
            message = self._prepare(Regular.get_function_doc_name(self.long_tap.__doc__))
            ActionChains(self._handler).click_and_hold(self._element_container[eid]).perform()
            time.sleep(sleep)
            ActionChains(self._handler).release(self._element_container[eid]).perform()
            execution_step_record(self._driver_id, message, str(f'{ sleep } 秒'))

        except BaseException as error:
            raise WebPlantElementFunctionRunError(self._driver_id, message, error)

    # ------------------------------
    def move_to(self, eid: int = 0):
        """
        [ 鼠标悬停 ]

        * 模拟一次作用于 WebElement 对象上的鼠标悬停操作.

        ---
        参数:
            eid { int, optional } : 指定元素下标, 默认为 0.

        """

        try:
            message = self._prepare(Regular.get_function_doc_name(self.move_to.__doc__))
            ActionChains(self._handler).move_to_element(self._element_container[eid]).perform()
            execution_step_record(self._driver_id, message)

        except BaseException as error:
            raise WebPlantElementFunctionRunError(self._driver_id, message, error)


# ----------------------------------
class _WebPlantElementMapInput(_WebPlantElementMapBase):
    """
    [ Web元素控件'输入框'相关函数 ]

    """

    # ------------------------------
    def send(self, key: str, eid: int = 0):
        """
        [ 输入内容 ]

        * 输入内容, 一般指向 input 标签组件.

        ---
        参数:
            key { str } : 内容文本.
            eid { int, optional } : 指定元素下标, 默认为 0.

        """

        try:
            message = self._prepare(Regular.get_function_doc_name(self.send.__doc__))
            self._element_container[eid].clear()
            self._element_container[eid].send_keys(key)
            execution_step_record(self._driver_id, message, key)

        except BaseException as error:
            raise WebPlantElementFunctionRunError(self._driver_id, message, error)

    # ------------------------------
    def clear(self, eid: int = 0):
        """
        [ 清除内容 ]

        * 清除内容, 一般指向 input 标签组件. 

        ---
        参数:
            eid { int, optional } : 指定元素下标, 默认为 0.

        """

        try:
            message = self._prepare(Regular.get_function_doc_name(self.clear.__doc__))
            self._element_container[eid].clear()
            execution_step_record(self._driver_id, message)

        except BaseException as error:
            raise WebPlantElementFunctionRunError(self._driver_id, message, error)


# ----------------------------------
class _WebPlantElementMapSelector(_WebPlantElementMapBase):
    """
    [ Web元素控件'下拉框'相关函数 ]

    """

    _select_handle: Select
    """ [ 下拉框处理器 ] """

    # ------------------------------
    def get_option_items(self, eid: int = 0) -> list:
        """
        [ 获取下拉框中所有的选项文本 ]

        ---
        参数:
            eid { int, optional } : 指定元素下标, 默认为 0.

        ---
        返回:
            list : 包含下拉框中所有的选项文本的列表.

        """

        try:
            message = self._prepare(Regular.get_function_doc_name(self.get_option_items.__doc__))
            self._init_select_handle(eid)

            texts = []
            option_list = self._select_handle.options
            for i in option_list:
                texts.append(i.text)

            execution_step_record(self._driver_id, message, str(texts))

            return texts

        except BaseException as error:
            raise WebPlantElementFunctionRunError(self._driver_id, message, error)

    # ------------------------------
    def get_option_select_items(self, eid: int = 0) -> list:
        """
        [ 获取下拉框中所有的已选中选项文本 ]

        ---
        参数:
            eid { int, optional } : 指定元素下标, 默认为 0.

        ---
        返回:
            list : 包含下拉框中所有的已选中选项文本的列表.

        """

        try:
            message = self._prepare(Regular.get_function_doc_name(self.get_option_select_items.__doc__))
            self._init_select_handle(eid)

            texts = []
            option_list = self._select_handle.all_selected_options
            for i in option_list:
                texts.append(i.text)

            execution_step_record(self._driver_id, message, str(texts))

            return texts

        except BaseException as error:
            raise WebPlantElementFunctionRunError(self._driver_id, message, error)

    # ------------------------------
    def get_option_current_item(self, eid: int = 0) -> str:
        """
        [ 获取下拉框中当前选择的首个选项文本 ]

        ---
        参数:
            eid { int, optional } : 指定元素下标, 默认为 0.

        ---
        返回:
            str : 下拉框中当前选择的首个选项文本.

        """

        try:
            message = self._prepare(Regular.get_function_doc_name(self.get_option_current_item.__doc__))
            self._init_select_handle(eid)
            text = self._select_handle.first_selected_option.text
            execution_step_record(self._driver_id, message, text)

            return text

        except BaseException as error:
            raise WebPlantElementFunctionRunError(self._driver_id, message, error)

    # ------------------------------
    def select_by_index(self, index: int, eid: int = 0):
        """
        [ 下拉框中通过下标选择选项 ]

        ---
        参数:
            index { int } : 需要选中的选项下标.
            eid { int, optional } : 指定元素下标, 默认为 0.

        """

        try:
            message = self._prepare(Regular.get_function_doc_name(self.select_by_index.__doc__))
            self._init_select_handle(eid)
            self._select_handle.select_by_index(index)
            select_option_text = self._select_handle.first_selected_option.text
            execution_step_record(self._driver_id, message, select_option_text)

        except BaseException as error:
            raise WebPlantElementFunctionRunError(self._driver_id, message, error)

    # ------------------------------
    def select_by_value(self, value: str, eid: int = 0):
        """
        [ 下拉框中通过选项标签属性值选择 ]

        ---
        参数:
            value { str } : 需要选中的选项标签属性值.
            eid { int, optional } : 指定元素下标, 默认为 0.

        """

        try:
            message = self._prepare(Regular.get_function_doc_name(self.select_by_value.__doc__))
            self._init_select_handle(eid)
            self._select_handle.select_by_value(value)
            select_option_text = self._select_handle.first_selected_option.text
            execution_step_record(self._driver_id, message, select_option_text)

        except BaseException as error:
            raise WebPlantElementFunctionRunError(self._driver_id, message, error)

    # ------------------------------
    def select_by_text(self, text: str, eid: int = 0):
        """
        [ 下拉框中通过选项的文本选择 ]

        ---
        参数:
            text { str } : 需要选中的选项文本.
            eid { int, optional } : 指定元素下标, 默认为 0.

        """

        try:
            message = self._prepare(Regular.get_function_doc_name(self.select_by_text.__doc__))
            self._init_select_handle(eid)
            self._select_handle.select_by_visible_text(text)
            select_option_text = self._select_handle.first_selected_option.text
            execution_step_record(self._driver_id, message, select_option_text)

        except BaseException as error:
            raise WebPlantElementFunctionRunError(self._driver_id, message, error)

    # ------------------------------
    def deselect_all(self, eid: int = 0):
        """
        [ 下拉框中取消所有的已选中选项 ]

        ---
        参数:
            eid { int, optional } : 指定元素下标, 默认为 0.

        """

        try:
            message = self._prepare(Regular.get_function_doc_name(self.deselect_all.__doc__))
            self._init_select_handle(eid)
            self._select_handle.deselect_all()
            execution_step_record(self._driver_id, message)

        except BaseException as error:
            raise WebPlantElementFunctionRunError(self._driver_id, message, error)

    # ------------------------------
    def deselect_by_index(self, index: int, eid: int = 0):
        """
        [ 下拉框中通过下标取消已选中选项 ]

        ---
        参数:
            index { int } : 需要取消选中的选项下标.
            eid { int, optional } : 指定元素下标, 默认为 0.

        """

        try:
            message = self._prepare(Regular.get_function_doc_name(self.deselect_by_index.__doc__))
            self._init_select_handle(eid)
            self._select_handle.select_by_index(index)
            select_option_text =self._select_handle.first_selected_option.text 
            self._select_handle.deselect_by_index(index)

            execution_step_record(self._driver_id, message, select_option_text)

        except BaseException as error:
            raise WebPlantElementFunctionRunError(self._driver_id, message, error)

    # ------------------------------
    def deselect_by_value(self, value: str, eid: int = 0):
        """
        [ 下拉框中通过选项标签属性值取消已选中选项 ]

        ---
        参数:
            value { str } : 需要取消选中的选项标签属性值.
            eid { int, optional } : 指定元素下标, 默认为 0.

        """

        try:
            message = self._prepare(Regular.get_function_doc_name(self.deselect_by_value.__doc__))
            self._init_select_handle(eid)
            self._select_handle.select_by_value(value)
            select_option_text =self._select_handle.first_selected_option.text 
            self._select_handle.deselect_by_value(value)

            execution_step_record(self._driver_id, message, select_option_text)

        except BaseException as error:
            raise WebPlantElementFunctionRunError(self._driver_id, message, error)

    # ------------------------------
    def deselect_by_text(self, text: str, eid: int = 0):
        """
        [ 下拉框中通过选项标签属性值取消已选中选项 ]

        ---
        参数:
            text { str } : 需要取消选中的选项的文本.
            eid { int, optional } : 指定元素下标, 默认为 0.

        """

        try:
            message = self._prepare(Regular.get_function_doc_name(self.deselect_by_text.__doc__))
            self._init_select_handle(eid)
            self._select_handle.select_by_visible_text(text)
            select_option_text =self._select_handle.first_selected_option.text 
            self._select_handle.deselect_by_visible_text(text)

            execution_step_record(self._driver_id, message, select_option_text)

        except BaseException as error:
            raise WebPlantElementFunctionRunError(self._driver_id, message, error)

    # ------------------------------
    def _init_select_handle(self, eid):
        """
        [ 初始化下拉框处理器 ]

        ---
        参数:
            eid { int, optional } : 指定元素下标, 默认为 0.

        """

        self._select_handle = Select(self._element_container[eid])


# ----------------------------------
def execution_step_record(driver_id: int, message: str, join_message: str = None, level: str = 'info'):
    """
    [ 执行步骤记录 ]

    * 执行时, 将同步将信息传递给日志服务与用例记录器.

    ---
    参数:
        driver_id { int } : 核心ID.
        message { str } : 记录信息.
        join_message { str } : 拼接信息.

    """

    if join_message:
        message = ''.join([message, ' >> ', f'[{ str(join_message) }]'])

    ACTION_RECORDER.write(driver_id, message)

    if CONFIG.debug:
        LOG.map(level, f'Driver[{ driver_id }] >> { message }')


# ----------------------------------
class WebPlantElementFunctionRunError(Exception):
    """ [ Web元素控件函数执行异常 ] """
    
    def __init__(self, driver_id: int, message: str, error: Any):
        self.message = f'error :: { message } :: { error } !'
        self.err_name = Regular.get_function_doc_name(self.__doc__)
        execution_step_record(driver_id, message, level='error')
        
    def __str__(self) -> str:
        return ''.join([self.err_name, ' :: ', self.message])
